/*-
# X-BASED TRIANGLES
#
#  Triangles.c
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Triangles */

#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "TrianglesP.h"

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif
#ifndef LOGFILE
#define LOGFILE "triangles.log"
#endif

static void InitializeTriangles(Widget request, Widget renew);
static void ExposeTriangles(Widget renew, XEvent * event, Region region);
static void ResizeTriangles(TrianglesWidget w);
static void DestroyTriangles(Widget old);
static Boolean SetValuesTriangles(Widget current, Widget request, Widget renew);
static void QuitTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void HideTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveTrianglesTl(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveTrianglesTr(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveTrianglesLeft(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveRightTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveTrianglesBl(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void MoveTrianglesBr(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void SelectTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void ReleaseTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeTrianglesMaybe(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeTriangles2(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void GetTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void WriteTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void ClearTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void UndoTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void SolveTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void IncrementTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void DecrementTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void EnterTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static void LeaveTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs);
static int  MoveTriangles(TrianglesWidget w, int direction);

static int  PositionToTile(TrianglesWidget w, int x, int y, int *row, int *trbl, int *tlbr);
static int  MovableTile(TrianglesWidget w);
#ifdef RHOMBUS
static int RhombusType(TrianglesWidget w);
#endif
static void SelectTiles(TrianglesWidget w);
static void SetAllColors(TrianglesWidget w, Boolean init);

static void CheckTiles(TrianglesWidget w);
static void ResetTiles(TrianglesWidget w);
static void ResizeTiles(TrianglesWidget w);
static void MoveNoTiles(TrianglesWidget w);
static int  MoveTilesDir(TrianglesWidget w, int direction, int fast);
static void GetTiles(TrianglesWidget w);
static void WriteTiles(TrianglesWidget w);
static void ClearTiles(TrianglesWidget w);
static void UndoTiles(TrianglesWidget w);
static void SolveTiles(TrianglesWidget w);
static void RandomizeTiles(TrianglesWidget w);
static void IncrementTiles(TrianglesWidget w);
static void DecrementTiles(TrianglesWidget w);
static void MoveTiles(TrianglesWidget w, int from, int orient, int fast);
static int  ExchangeTiles(TrianglesWidget w, int pos1, int pos2);
static int  TileNFromSpace(TrianglesWidget w, int n, int rowType, int orient, int direction);
static void DrawFrame(TrianglesWidget w, GC gc);
static void DrawTile(TrianglesWidget w, int pos, int orient, Boolean blank,
	Boolean erase, int pressedOffset, int offsetX, int offsetY);
static Pixel brighter(TrianglesWidget w, Pixel pixel);
static Pixel darker(TrianglesWidget w, Pixel pixel);
static void drawShadow(TrianglesWidget w, GC gc, int startX, int startY,
	XPoint * list, Boolean orient);
static void fill3DTriangle(TrianglesWidget w, GC gc, GC darkerGC, GC brighterGC,
	XPoint * list, Boolean raised, Boolean orient);
static void DiscreteMoves(TrianglesWidget w, int dir, int fast);
#ifdef ANIMATE
static int CountTiles(TrianglesWidget w, int * dir, int * rowType);
static void AnimateSlide(TrianglesWidget w, int dir, int fast, Bool logMoves);
#endif
static Boolean bresenhamLine(TrianglesWidget w, int x0, int y0, int x1, int y1,
	int * pixX, int * pixY, int * pixNo);
static void Cartesian(TrianglesWidget w, int pos, int orient, int * x, int * y);
static int  ToOrient(int row, int trbl, int tlbr);
static int  ToPosition(int row, int trbl, int tlbr);
static int  Sqrt(int i);
static int  int2String(char *buf, int number, int base);

static char defaultTranslationsTriangles[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>Home: MoveTl()\n\
 <KeyPress>KP_7: MoveTl()\n\
 <KeyPress>R7: MoveTl()\n\
 <KeyPress>Prior: MoveTr()\n\
 <KeyPress>KP_9: MoveTr()\n\
 <KeyPress>R9: MoveTr()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <KeyPress>End: MoveBl()\n\
 <KeyPress>KP_1: MoveBl()\n\
 <KeyPress>R13: MoveBl()\n\
 <KeyPress>Next: MoveBr()\n\
 <KeyPress>KP_3: MoveBr()\n\
 <KeyPress>R15: MoveBr()\n\
 <Btn1Down>: Select()\n\
 <Btn1Up>: Release()\n\
 <KeyPress>r: Randomize()\n\
 <Btn3Down>: RandomizeMaybe()\n\
 <Btn3Down>(2+): Randomize2()\n\
 <KeyPress>g: Get()\n\
 <KeyPress>w: Write()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>u: Undo()\n\
 <KeyPress>s: Solve()\n\
 <KeyPress>i: Increment()\n\
 <KeyPress>d: Decrement()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListTriangles[] =
{
	{(char *) "Quit", (XtActionProc) QuitTriangles},
	{(char *) "Hide", (XtActionProc) HideTriangles},
	{(char *) "MoveTl", (XtActionProc) MoveTrianglesTl},
	{(char *) "MoveTr", (XtActionProc) MoveTrianglesTr},
	{(char *) "MoveLeft", (XtActionProc) MoveTrianglesLeft},
	{(char *) "MoveRight", (XtActionProc) MoveRightTriangles},
	{(char *) "MoveBl", (XtActionProc) MoveTrianglesBl},
	{(char *) "MoveBr", (XtActionProc) MoveTrianglesBr},
	{(char *) "Select", (XtActionProc) SelectTriangles},
	{(char *) "Release", (XtActionProc) ReleaseTriangles},
	{(char *) "Randomize", (XtActionProc) RandomizeTriangles},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizeTrianglesMaybe},
	{(char *) "Randomize2", (XtActionProc) RandomizeTriangles2},
	{(char *) "Get", (XtActionProc) GetTriangles},
	{(char *) "Write", (XtActionProc) WriteTriangles},
	{(char *) "Clear", (XtActionProc) ClearTriangles},
	{(char *) "Undo", (XtActionProc) UndoTriangles},
	{(char *) "Solve", (XtActionProc) SolveTriangles},
	{(char *) "Increment", (XtActionProc) IncrementTriangles},
	{(char *) "Decrement", (XtActionProc) DecrementTriangles},
	{(char *) "Enter", (XtActionProc) EnterTriangles},
	{(char *) "Leave", (XtActionProc) LeaveTriangles}
};

static XtResource resourcesTriangles[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(TrianglesWidget, triangles.username),
	 XtRString, (caddr_t) "nobody"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(TrianglesWidget, triangles.foreground),
	 XtRString, (caddr_t) "Cyan" /*XtDefaultForeground*/},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(TrianglesWidget, triangles.background),
	 XtRString, (caddr_t) "NavyBlue" /*XtDefaultBackground*/},
	{XtNtileColor, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(TrianglesWidget, triangles.tileColor),
	 XtRString, (caddr_t) "LightGray" /*XtDefaultForeground*/},
	{XtNtileBorder, XtCColor, XtRPixel, sizeof (Pixel),
	 XtOffset(TrianglesWidget, triangles.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultBackground*/},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(TrianglesWidget, core.width),
	 XtRString, (caddr_t) "200"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(TrianglesWidget, core.height),
	 XtRString, (caddr_t) "173"},
	{XtNsize, XtCSize, XtRInt, sizeof (int),
	 XtOffset(TrianglesWidget, triangles.size),
	 XtRString, (caddr_t) "4"},	/* DEFAULTTRIS */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(TrianglesWidget, triangles.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(TrianglesWidget, triangles.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmoveDelay, XtCMoveDelay, XtRInt, sizeof (int),
	 XtOffset(TrianglesWidget, triangles.moveDelay),
	 XtRString, (caddr_t) "10"},
	{XtNbase, XtCBase, XtRInt, sizeof (int),
	 XtOffset(TrianglesWidget, triangles.base),
	 XtRString, (caddr_t) "10"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TrianglesWidget, triangles.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNcheat, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(TrianglesWidget, triangles.cheat),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(TrianglesWidget, triangles.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(TrianglesWidget, triangles.select),
	 XtRCallback, (caddr_t) NULL}
};

TrianglesClassRec trianglesClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Triangles",	/* class name */
		sizeof (TrianglesRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeTriangles,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListTriangles,	/* actions */
		XtNumber(actionsListTriangles),		/* num actions */
		resourcesTriangles,	/* resources */
		XtNumber(resourcesTriangles),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyTriangles,	/* destroy */
		(XtWidgetProc) ResizeTriangles,		/* resize */
		(XtExposeProc) ExposeTriangles,		/* expose */
		(XtSetValuesFunc) SetValuesTriangles,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsTriangles,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass trianglesWidgetClass = (WidgetClass) & trianglesClassRec;

static XPoint triangleUnit[MAXORIENT][ROWTYPES + 1] =
{
	{
		{0, 0},
		{-1, -1},
		{2, 0},
		{-1, 1}},
	{
		{0, 0},
		{1, 1},
		{-2, 0},
		{1, -1}}
};
static XPoint triangleList[MAXORIENT][ROWTYPES + 1];
#ifdef RHOMBUS
static XPoint rhombusUnit[ROWTYPES][ROWTYPES + 2] =
{
	{
		{0, 0},
		{-2, 0},
		{1, 1},
		{2, 0},
		{-1, -1}},
	{
		{0, 0},
		{-1, 1},
		{2, 0},
		{1, -1},
		{-2, -0}},
	{
		{0, 0},
		{-1, 1},
		{1, 1},
		{1, -1},
		{-1, -1}}
};
static XPoint rhombusList[ROWTYPES][ROWTYPES + 2];
#endif

#ifndef HAVE_USLEEP
#if !defined( VMS ) || defined( XVMSUTILS ) ||  ( __VMS_VER >= 70000000 )
#ifdef USE_XVMSUTILS
#include <X11/unix_time.h>
#endif
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#endif
#endif
#if defined(SYSV) || defined(SVR4)
#ifdef LESS_THAN_AIX3_2
#include <sys/poll.h>
#else /* !LESS_THAN_AIX3_2 */
#include <poll.h>
#endif /* !LESS_THAN_AIX3_2 */
#endif /* defined(SYSV) || defined(SVR4) */

/* not static in case usleep found in system include */
int
usleep(unsigned int usec)
{
#if (defined (SYSV) || defined(SVR4)) && !defined(__hpux)
#if defined(HAVE_NANOSLEEP)
	{
		struct timespec rqt;

		rqt.tv_nsec = 1000 * (usec % (unsigned int) 1000000);
		rqt.tv_sec = usec / (unsigned int) 1000000;
		return nanosleep(&rqt, NULL);
	}
#else
	(void) poll(
#if defined(__cplusplus) || defined(c_plusplus)
		(pollfd *) /* guess */
#else
		(void *)
#endif
		0, (int) 0, usec / 1000);	/* ms resolution */
#endif
#else
#ifdef VMS
	long        timadr[2];

	if (usec != 0) {
		timadr[0] = -usec * 10;
		timadr[1] = -1;

		sys$setimr(4, &timadr, 0, 0, 0);
		sys$waitfr(4);
	}
#else
	struct timeval time_out;

#if  0
	/* (!defined(AIXV3) && !defined(__hpux)) */
	extern int  select(int, fd_set *, fd_set *, fd_set *, struct timeval *);

#endif

	time_out.tv_usec = usec % (unsigned int) 1000000;
	time_out.tv_sec = usec / (unsigned int) 1000000;
	(void) select(0, (void *) 0, (void *) 0, (void *) 0, &time_out);
#endif
#endif
	return 0;
}
#endif

static void
Sleep(unsigned int cMilliseconds)
{
	(void) usleep(cMilliseconds * 1000);
}

void
intCat(char ** string, const char * var1, const int var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + 21)))
        	XtError("Not enough memory, exiting.");
	(void) sprintf(*string, "%s%d", var1, var2);
}

void
stringCat(char ** string, const char * var1, const char * var2)
{
	if (!(*string = (char *) malloc(strlen(var1) + strlen(var2) + 1)))
        	XtError("Not enough memory, exiting.");
	(void) sprintf(*string, "%s%s", var1, var2);
}

static void
InitializeTriangles(Widget request, Widget renew)
{
	TrianglesWidget w = (TrianglesWidget) renew;

	w->triangles.tileOfPosition = NULL;
	CheckTiles(w);
	InitMoves();
	ResetTiles(w);
	w->triangles.numSlices = ((w->triangles.moveDelay < MAXSLICES) ?
		w->triangles.moveDelay + 1 : MAXSLICES);
	(void) SRAND(getpid());
	w->triangles.cheat = False;
	w->triangles.depth = DefaultDepthOfScreen(XtScreen(w));
	SetAllColors(w, True);
	ResizeTriangles(w);
}

static void
DestroyTriangles(Widget old)
{
	TrianglesWidget w = (TrianglesWidget) old;

	XtReleaseGC(old, w->triangles.tileGC);
	XtReleaseGC(old, w->triangles.tileBrighterGC);
	XtReleaseGC(old, w->triangles.tileDarkerGC);
	XtReleaseGC(old, w->triangles.borderGC);
	XtReleaseGC(old, w->triangles.puzzleGC);
	XtReleaseGC(old, w->triangles.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->triangles.select);
}

static void
ResizeTriangles(TrianglesWidget w)
{
	double      sqrt_3 = 1.73205080756887729352744634150587237;
	XPoint      tempSize;

	w->triangles.delta.x = 5;
	w->triangles.delta.y = 3;
	w->triangles.offset.x = MAX(((int) w->core.width -
			       w->triangles.delta.x) / w->triangles.size, 0);
	w->triangles.offset.y = MAX(((int) w->core.height -
			   2 * w->triangles.delta.y) / w->triangles.size, 0);
	tempSize.y = (int) ((double) w->triangles.offset.x * sqrt_3 / 2.0);
	tempSize.x = (int) ((double) w->triangles.offset.y * 2.0 / sqrt_3);
	if (tempSize.y < w->triangles.offset.y)
		w->triangles.offset.y = tempSize.y;
	else			/* tempSize.x <=  w->triangles.wid */
		w->triangles.offset.x = tempSize.x;
	w->triangles.puzzleSize.x = w->triangles.offset.x * w->triangles.size +
		w->triangles.delta.x + 2;
	w->triangles.puzzleSize.y = w->triangles.offset.y * w->triangles.size +
		w->triangles.delta.y + 2;
	w->triangles.puzzleOffset.x = ((int) w->core.width -
				       w->triangles.puzzleSize.x + 2) / 2;
	w->triangles.puzzleOffset.y = ((int) w->core.height -
				       w->triangles.puzzleSize.y + 2) / 2;
	w->triangles.tileSize.x = MAX(w->triangles.offset.x - w->triangles.delta.x,
				      0);
	w->triangles.tileSize.y = MAX(w->triangles.offset.y - w->triangles.delta.y,
				      0);
	ResizeTiles(w);
}

static void
ExposeTriangles(Widget renew, XEvent * event, Region region)
{
	TrianglesWidget w = (TrianglesWidget) renew;

	if (w->core.visible) {
		XFillRectangle(XtDisplay(w), XtWindow(w),
			w->triangles.inverseGC,
			0, 0, w->core.width, w->core.height);
		DrawFrame(w, (w->triangles.focus) ? w->triangles.puzzleGC :
			w->triangles.inverseGC);
		DrawAllTiles(w);
	}
}

static      Boolean
SetValuesTriangles(Widget current, Widget request, Widget renew)
{
	TrianglesWidget c = (TrianglesWidget) current, w = (TrianglesWidget) renew;
	Boolean     redraw = FALSE;
	Boolean     redrawTiles = FALSE;

	CheckTiles(w);
	if (w->triangles.background != c->triangles.background ||
	    w->triangles.foreground != c->triangles.foreground ||
	    w->triangles.borderColor != c->triangles.borderColor ||
	    w->triangles.tileColor != c->triangles.tileColor ||
	    w->triangles.reverse != c->triangles.reverse ||
	    w->triangles.mono != c->triangles.mono) {
		SetAllColors(w, False);
		redrawTiles = TRUE;
	}
	if (w->triangles.size != c->triangles.size ||
	    w->triangles.base != c->triangles.base) {
		ResetTiles(w);
		ResizeTriangles(w);
		redraw = TRUE;
	} else if (w->triangles.offset.x != c->triangles.offset.x ||
		   w->triangles.offset.y != c->triangles.offset.y) {
		ResizeTriangles(w);
		redraw = TRUE;
	}
	if (w->triangles.moveDelay != c->triangles.moveDelay) {
		w->triangles.numSlices = ((w->triangles.moveDelay < MAXSLICES) ?
			w->triangles.moveDelay + 1 : MAXSLICES);
	}
	if (w->triangles.menu != -1) {
		switch (w->triangles.menu) {
		case 0:
			w->triangles.menu = -1;
			GetTiles(w);
			break;
		case 1:
			w->triangles.menu = -1;
			WriteTiles(w);
			break;
		case 3:
			w->triangles.menu = -1;
			ClearTiles(w);
			break;
		case 4:
			w->triangles.menu = -1;
			UndoTiles(w);
			break;
		case 5:
			w->triangles.menu = -1;
			RandomizeTiles(w);
			break;
		case 6:
			w->triangles.menu = -1;
			SolveTiles(w);
			break;
		case 7:
			w->triangles.menu = -1;
			IncrementTiles(w);
			break;
		case 8:
			w->triangles.menu = -1;
			DecrementTiles(w);
			break;
		default:
			w->triangles.menu = -1;
			break;
		}
	}
	if (redrawTiles && !redraw && XtIsRealized(renew) && renew->core.visible) {
		DrawFrame(c, c->triangles.inverseGC);
		if (w->triangles.focus)
			DrawFrame(w, w->triangles.puzzleGC);
		DrawAllTiles(w);
	}
	return (redraw);
}

static void
QuitTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
HideTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	trianglesCallbackStruct cb;

	cb.reason = TRIANGLES_HIDE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
SelectTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	int         pos, row, trbl, tlbr, rowType;

	pos = PositionToTile(w, event->xbutton.x, event->xbutton.y,
			     &row, &trbl, &tlbr);
	if (pos >= 0) {
		if (CheckSolved(w)) {
			MoveNoTiles(w);
			w->triangles.currentPosition = -1;
			return;
		}
		w->triangles.currentPosition = pos;
		w->triangles.currentRow[ROW] = row;
		w->triangles.currentRow[TRBL] = trbl;
		w->triangles.currentRow[TLBR] = tlbr;
		w->triangles.currentPositionOrient = ToOrient(row, trbl, tlbr);
		rowType = MovableTile(w);
		if (rowType < 0) {
			trianglesCallbackStruct cb;

			DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, rowType == TRIANGLES_SPACE, False, TRUE, 0, 0);
			XFlush(XtDisplay(w));
			Sleep(100);
			DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, True, True, TRUE, 0, 0);
			if (rowType != TRIANGLES_SPACE)
				DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, False, False, FALSE, 0, 0);
			cb.reason = rowType;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			w->triangles.currentPosition = -1;
			return;
		}
		DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, False, False, TRUE, 0, 0);
	} else
		w->triangles.currentPosition = -1;
}

static void
ReleaseTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	if (w->triangles.currentPosition < 0)
		return;
	DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, True, True, TRUE, 0, 0);
	DrawTile(w, w->triangles.currentPosition, w->triangles.currentPositionOrient, False, False, FALSE, 0, 0);
	SelectTiles(w);
	w->triangles.currentPosition = -1;
}

static void
RandomizeTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	RandomizeTiles(w);
}

static void
RandomizeTrianglesMaybe(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	if (!w->triangles.started)
		RandomizeTiles(w);
#ifdef HAVE_MOTIF
	else {
		trianglesCallbackStruct cb;

		cb.reason = TRIANGLES_RANDOMIZE_QUERY;
		XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
	}
#endif
}

static void
RandomizeTriangles2(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
#ifdef HAVE_MOTIF
	if (!w->triangles.started)
#endif
		RandomizeTiles(w);
}

static void
GetTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	GetTiles(w);
}

static void
WriteTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	WriteTiles(w);
}

static void
ClearTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	ClearTiles(w);
}

static void
UndoTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	UndoTiles(w);
}

static void
SolveTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	SolveTiles(w);
}

static void
IncrementTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	IncrementTiles(w);
}

static void
DecrementTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	DecrementTiles(w);
}

static void
EnterTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	w->triangles.focus = True;
	DrawFrame(w, w->triangles.puzzleGC);
}

static void
LeaveTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	w->triangles.focus = False;
	DrawFrame(w, w->triangles.inverseGC);
}

static void
MoveTrianglesTl(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, TL);
}

static void
MoveTrianglesTr(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, TR);
}

static void
MoveTrianglesLeft(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, LEFT);
}

static void
MoveRightTriangles(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, RIGHT);
}

static void
MoveTrianglesBl(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, BL);
}

static void
MoveTrianglesBr(TrianglesWidget w, XEvent * event, char **args, int nArgs)
{
	(void) MoveTriangles(w, BR);
}

static int
MoveTriangles(TrianglesWidget w, int direction)
{
	trianglesCallbackStruct cb;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return FALSE;
	}
	if (!MoveTrianglesDir(w, direction, False)) {
		cb.reason = TRIANGLES_BLOCKED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		return FALSE;
	}
	if (CheckSolved(w)) {
		cb.reason = TRIANGLES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return TRUE;
}

int
MoveTrianglesDir(TrianglesWidget w, int direction, int fast)
{
	trianglesCallbackStruct cb;

	if (MoveTilesDir(w, direction, fast)) {
		cb.reason = TRIANGLES_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		PutMove(direction);
		return TRUE;
	}
	return FALSE;
}

static int
PositionToTile(TrianglesWidget w, int x, int y, int *row, int *trbl, int *tlbr)
{
	int         sumX, sumY, sumX2;

	sumX = w->triangles.size * w->triangles.offset.x + w->triangles.delta.x;
	sumY = w->triangles.size * w->triangles.offset.y + w->triangles.delta.y + 1;
	sumX2 = sumX / 2;
	x -= w->triangles.puzzleOffset.x;
	y -= w->triangles.puzzleOffset.y;
	if (x * (sumY + w->triangles.delta.y) + y * (sumX2 + 1) <
	    (sumX2 + 1) * (sumY + w->triangles.delta.y) ||
	    x * (sumY + w->triangles.delta.y) - y * (sumX2 - 1) >
	    (sumX2 - 1) * (sumY + w->triangles.delta.y) ||
	    y > sumY - w->triangles.delta.y)
		return -1;
	*row = (y - w->triangles.delta.y) / w->triangles.offset.y;
	*trbl = (x - sumX2 - 1 + *row * w->triangles.offset.x / 2) /
		w->triangles.offset.x;
	*trbl += ((x - (*trbl + 1) * w->triangles.offset.x) *
		  (sumY + w->triangles.delta.y) + y * (sumX2 + 1)) /
		((sumX2 + 1) * (sumY + w->triangles.delta.y));
	*tlbr = (-x + sumX2 - 1 + *row * w->triangles.offset.x / 2) /
		w->triangles.offset.x;
	*tlbr += 1 + ((-x - (*tlbr + 1) * w->triangles.offset.x) *
		      (sumY + w->triangles.delta.y) + y * (sumX2 - 1)) /
		((sumX2 - 1) * (sumY + w->triangles.delta.y));
	if (*row >= 0 && *trbl >= 0 && *tlbr >= 0 && *row < w->triangles.size &&
	    *trbl < w->triangles.size && *tlbr < w->triangles.size) {
		return ToPosition(*row, *trbl, *tlbr);
	} else
		return -1;
}

static int
MovableTile(TrianglesWidget w)
{
	int         rowType = TRIANGLES_BLOCKED, l;

	/* Are the spaces in a "row" with the mouse click?
	   (If two, then one clicked on a space). */
	for (l = 0; l < ROWTYPES; l++) {
		if (w->triangles.currentRow[l] == w->triangles.spaceRow[DOWN][l] &&
		w->triangles.currentRow[l] == w->triangles.spaceRow[UP][l]) {
			if (rowType == TRIANGLES_BLOCKED) {
				rowType = l;
			} else {
				return TRIANGLES_SPACE;
			}
		}
	}
	return rowType;
}

#ifdef RHOMBUS
/* Rhombus type is determined by its narrowist "width" to reuse constants */
static int
RhombusType(TrianglesWidget w)
{
	if (w->triangles.spaceRow[0][ROW] != w->triangles.spaceRow[1][ROW])
		return ROW;
	if (w->triangles.spaceRow[0][TRBL] != w->triangles.spaceRow[1][TRBL])
		return TRBL;
	if (w->triangles.spaceRow[0][TLBR] != w->triangles.spaceRow[1][TLBR])
		return TLBR;
	XtWarning("Rhombus Type: unknown");
	return -1;
}
#endif

static void
SelectTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	if (w->triangles.currentPosition <
	    w->triangles.spacePosition[w->triangles.currentPositionOrient]) {
#ifdef ANIMATE
		if (w->triangles.moveDelay > 0) {
			AnimateSlide(w, DOWN, NORMAL, True);
		} else
#endif
			while (w->triangles.currentPosition <
			       w->triangles.spacePosition[w->triangles.currentPositionOrient]) {

				DiscreteMoves(w, DOWN, NORMAL);
			}
	} else {		/*w->triangles.currentPosition >
				   w->triangles.spacePosition[w->triangles.currentPositionOrient] */
#ifdef ANIMATE
		if (w->triangles.moveDelay > 0) {
			AnimateSlide(w, UP, NORMAL, True);
		} else
#endif
			while (w->triangles.currentPosition >
			       w->triangles.spacePosition[w->triangles.currentPositionOrient]) {
				DiscreteMoves(w, UP, NORMAL);
			}
	}
	if (CheckSolved(w)) {
		cb.reason = TRIANGLES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

#define FACTOR 0.7
#define MAXINTENSITY 0xFFFF

static Pixel
brighter(TrianglesWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(TrianglesWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w), DefaultColormapOfScreen(XtScreen(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
SetAllColors(TrianglesWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;

	valueMask = GCForeground | GCBackground;
	if (w->triangles.reverse) {
		values.foreground = w->triangles.foreground;
		values.background = w->triangles.background;
	} else {
		values.foreground = w->triangles.background;
		values.background = w->triangles.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.inverseGC);
	w->triangles.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->triangles.reverse) {
		values.foreground = w->triangles.background;
		values.background = w->triangles.foreground;
	} else {
		values.foreground = w->triangles.foreground;
		values.background = w->triangles.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.puzzleGC);
	w->triangles.puzzleGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->triangles.depth < 2 || w->triangles.mono) {
		if (w->triangles.reverse) {
			values.foreground = w->triangles.background;
			values.background = w->triangles.foreground;
		} else {
			values.foreground = w->triangles.foreground;
			values.background = w->triangles.background;
		}
	} else {
		values.foreground = w->triangles.tileColor;
		values.background = w->triangles.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.tileGC);
	w->triangles.tileGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->triangles.depth >= 2 && !w->triangles.mono) {
		values.foreground = brighter(w, w->triangles.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.tileBrighterGC);
	w->triangles.tileBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->triangles.depth >= 2 && !w->triangles.mono) {
		values.foreground = darker(w, w->triangles.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.tileDarkerGC);
	w->triangles.tileDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->triangles.depth < 2 || w->triangles.mono) {
		if (w->triangles.reverse) {
			values.foreground = w->triangles.foreground;
			values.background = w->triangles.background;
		} else {
			values.foreground = w->triangles.background;
			values.background = w->triangles.foreground;
		}
	} else {
		values.foreground = w->triangles.borderColor;
		values.background = w->triangles.tileColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->triangles.borderGC);
	w->triangles.borderGC = XtGetGC((Widget) w, valueMask, &values);
}

static void
CheckTiles(TrianglesWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->triangles.size < MINTRIANGLES) {
		intCat(&buf1,
			"Number of triangles on an edge out of bounds, use at least ",
			MINTRIANGLES);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTTRIANGLES);
		free(buf2);
		XtWarning(buf1);
		free(buf1);
		w->triangles.size = DEFAULTTRIANGLES;
	}
	if (w->triangles.moveDelay < 0) {
		intCat(&buf1, "Delay can not be negative (",
			w->triangles.moveDelay);
		stringCat(&buf2, buf1, "), taking absolute value");
		free(buf1);
		XtWarning(buf2);
		free(buf2);
		w->triangles.moveDelay = -w->triangles.moveDelay;
	}
	if (w->triangles.base < MINBASE  || w->triangles.base > MAXBASE) {
		intCat(&buf1, "Base out of bounds, use ", MINBASE);
		stringCat(&buf2, buf1, "..");
		free(buf1);
		intCat(&buf1, buf2, MAXBASE);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULTBASE);
		free(buf2);
		XtWarning(buf1);
		free(buf1);
		w->triangles.base = DEFAULTBASE;
	}

}

static void
ResetTiles(TrianglesWidget w)
{
	int         i;

	w->triangles.sizeSize = w->triangles.size * w->triangles.size;
	if (w->triangles.tileOfPosition)
		(void) free((void *) w->triangles.tileOfPosition);
	if (!(w->triangles.tileOfPosition = (int *)
	      malloc(sizeof (int) * w->triangles.sizeSize)))
		            XtError("Not enough memory, exiting.");

	if (startPosition)
		(void) free((void *) startPosition);
	if (!(startPosition = (int *)
	      malloc(sizeof (int) * w->triangles.sizeSize)))
		            XtError("Not enough memory, exiting.");

	w->triangles.spacePosition[UP] = w->triangles.sizeSize - 1;
	w->triangles.spaceRow[UP][TRBL] = w->triangles.size - 1;	/*i */
	w->triangles.spaceRow[UP][TLBR] = 0;	/*j */
	w->triangles.spaceRow[UP][ROW] = w->triangles.size - 1;		/*k */
	if (w->triangles.size > 1) {
		w->triangles.spacePosition[DOWN] = w->triangles.sizeSize - 2;
		w->triangles.spaceRow[DOWN][TRBL] = w->triangles.size - 2;	/*i */
		w->triangles.spaceRow[DOWN][TLBR] = 0;	/*j */
		w->triangles.spaceRow[DOWN][ROW] = w->triangles.size - 1;	/*k */
		w->triangles.tileOfPosition[w->triangles.sizeSize - 2] = -1;
	}
	w->triangles.tileOfPosition[w->triangles.sizeSize - 1] = 0;
	for (i = 1; i < w->triangles.sizeSize - 1; i++)
		w->triangles.tileOfPosition[i - 1] = i;
	FlushMoves(w);
	w->triangles.currentPosition = -1;
	w->triangles.started = FALSE;
}

static void
ResizeTiles(TrianglesWidget w)
{
	int         i, j;

	for (j = 0; j < MAXORIENT; j++)
		for (i = 0; i <= ROWTYPES; i++) {
			triangleList[j][i].x = (w->triangles.tileSize.x / 2) *
				triangleUnit[j][i].x;
			triangleList[j][i].y = w->triangles.tileSize.y *
				triangleUnit[j][i].y;
		}
#ifdef RHOMBUS
	for (j = 0; j < ROWTYPES; j++)
		for (i = 0; i <= ROWTYPES + 1; i++) {
			rhombusList[j][i].x = ((w->triangles.tileSize.x + w->triangles.delta.x) / 2 ) *
				rhombusUnit[j][i].x;
			rhombusList[j][i].y = (w->triangles.tileSize.y + w->triangles.delta.y - 1) *
				rhombusUnit[j][i].y;
		}
#endif
	w->triangles.digitOffset.x = 3;
	w->triangles.digitOffset.y = 2;
}

static void
MoveNoTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	cb.reason = TRIANGLES_IGNORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static int
MoveTilesDir(TrianglesWidget w, int direction, int fast)
{
	int         orient;

	switch (direction) {
		case TR:
			if (w->triangles.spaceRow[UP][TRBL] == w->triangles.spaceRow[DOWN][TRBL]
			    && w->triangles.spaceRow[UP][ROW] != w->triangles.size - 1) {
				orient = (w->triangles.spacePosition[UP] + 2 >
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, TRBL, orient, DOWN),
					!orient, fast);
				return TRUE;
			}
			break;
		case RIGHT:
			if (w->triangles.spaceRow[UP][ROW] == w->triangles.spaceRow[DOWN][ROW]
			    && w->triangles.spaceRow[UP][TRBL] != 0) {
				orient = (w->triangles.spacePosition[UP] <
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, ROW, orient, UP),
					!orient, fast);
				return TRUE;
			}
			break;
		case BR:
			if (w->triangles.spaceRow[UP][TLBR] == w->triangles.spaceRow[DOWN][TLBR]
			    && w->triangles.spaceRow[UP][TRBL] != 0) {
				orient = (w->triangles.spacePosition[UP] <
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, TLBR, orient, UP),
					!orient, fast);
				return TRUE;
			}
			break;
		case BL:
			if (w->triangles.spaceRow[UP][TRBL] == w->triangles.spaceRow[DOWN][TRBL]
			    && w->triangles.spaceRow[UP][TLBR] != 0) {
				orient = (w->triangles.spacePosition[UP] + 1 <
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, TRBL, orient, UP),
					!orient, fast);
				return TRUE;
			}
			break;
		case LEFT:
			if (w->triangles.spaceRow[UP][ROW] == w->triangles.spaceRow[DOWN][ROW]
			    && w->triangles.spaceRow[UP][TLBR] != 0) {
				orient = (w->triangles.spacePosition[UP] >
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, ROW, orient, DOWN),
					!orient, fast);
				return TRUE;
			}
			break;
		case TL:
			if (w->triangles.spaceRow[UP][TLBR] == w->triangles.spaceRow[DOWN][TLBR]
			    && w->triangles.spaceRow[UP][ROW] != w->triangles.size - 1) {
				orient = (w->triangles.spacePosition[UP] >
				w->triangles.spacePosition[DOWN]) ? UP : DOWN;
				MoveTiles(w, TileNFromSpace(w, 1, TLBR, orient, DOWN),
					!orient, fast);
				return TRUE;
			}
			break;
		default:
			{
				char *buf;

				intCat(&buf, "MoveTilesDir: direction ",
					direction);
				XtWarning(buf);
				free(buf);
			}
	}
	return FALSE;
}

static void
GetTiles(TrianglesWidget w)
{
	FILE       *fp;
	int         c, i, size, moves;
	trianglesCallbackStruct cb;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "r")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "r")) == NULL) {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not read (get) ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
		}
#endif
	}
	FlushMoves(w);
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &size);
	if (size >= MINTRIANGLES) {
		for (i = w->triangles.size; i < size; i++) {
			cb.reason = TRIANGLES_INC;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
		for (i = w->triangles.size; i > size; i--) {
			cb.reason = TRIANGLES_DEC;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	} else {
		stringCat(&buf1, name, " corrupted: size ");
		intCat(&buf2, buf1, size);
		free(buf1);
		stringCat(&buf1, buf2, " should be between ");
		free(buf2);
		intCat(&buf2, buf1, MINTRIANGLES);
		free(buf1);
		stringCat(&buf1, buf2, " and MAXINT");
		free(buf2);
		XtWarning(buf1);
		free(buf1);
	}
	while ((c = getc(fp)) != EOF && c != SYMBOL);
	(void) fscanf(fp, "%d", &moves);
	ScanStartPosition(fp, w);
	cb.reason = TRIANGLES_RESTORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	SetStartPosition(w);
	ScanMoves(fp, w, moves);
	(void) fclose(fp);
	(void) printf("%s: size %d, moves %d.\n", name, size, moves);
	free(lname);
	free(fname);
	w->triangles.cheat = True; /* Assume the worst. */
}

static void
WriteTiles(TrianglesWidget w)
{
	FILE       *fp;
	char *buf1 = NULL, *buf2 = NULL;
	char *fname, *lname, *name;

	stringCat(&buf1, CURRENTDELIM, LOGFILE);
	lname = buf1;
	stringCat(&buf1, LOGPATH, FINALDELIM);
	stringCat(&buf2, buf1, LOGFILE);
	free(buf1);
	fname = buf2;
	/* Try current directory first. */
	name = lname;
	if ((fp = fopen(name, "w")) == NULL) {
		name = fname;
		if ((fp = fopen(name, "w")) == NULL) {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, " or ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
			free(lname);
			free(fname);
			return;
		}
/* Probably annoying */
#if 0
		else {
			stringCat(&buf1, "Can not write to ", lname);
			stringCat(&buf2, buf1, ", falling back to ");
			free(buf1);
			stringCat(&buf1, buf2, fname);
			free(buf2);
			XtWarning(buf1);
			free(buf1);
		}
#endif
	}
	(void) fprintf(fp, "size%c %d\n", SYMBOL, w->triangles.size);
	(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
	PrintStartPosition(fp, w);
	PrintMoves(fp);
	(void) fclose(fp);
	(void) printf("Saved to %s.\n", name);
	free(lname);
	free(fname);
}

static void
ClearTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	if (w->triangles.currentPosition >= 0)
		return;
	ResetTiles(w);
	DrawAllTiles(w);
	cb.reason = TRIANGLES_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
UndoTiles(TrianglesWidget w)
{
	if (MadeMoves() && w->triangles.currentPosition < 0) {
		int         direction;

		GetMove(&direction);
		direction = (direction + (COORD / 2)) % COORD;
		if (MoveTilesDir(w, direction, DOUBLE)) {
			trianglesCallbackStruct cb;

			cb.reason = TRIANGLES_UNDO;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		} else {
			char *buf1, *buf2;

			intCat(&buf1, "Move ", direction);
			stringCat(&buf2, buf1, " can not be made");
			free(buf1);
			XtWarning(buf2);
			free(buf2);
		}
	}
}

static void
RandomizeTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	if (w->triangles.currentPosition >= 0)
		return;
	w->triangles.cheat = False;
	/* First interchange tiles but only with other tiles of the same
	   orientation */
	if (w->triangles.size > 2) {
		int         currentPos, randomPos, randomRow;
		int         currentOrient = UP, randomOrient = DOWN;
		int         step = 1, fin = 1;

		for (currentPos = 0; currentPos < w->triangles.sizeSize; currentPos++) {
			randomPos = currentPos;
			while (currentPos == randomPos || currentOrient != randomOrient) {
				randomPos = NRAND(w->triangles.sizeSize);
				randomRow = Row(randomPos);
				randomOrient = !((randomRow + TrBl(randomPos, randomRow) +
					TlBr(randomPos, randomRow)) % 2);
			}
			(void) ExchangeTiles(w, currentPos, randomPos);
			if (fin == currentPos + 1) {
				currentOrient = UP;
				step += 2;
				fin += step;
			} else
				currentOrient = !currentOrient;
		}
		DrawAllTiles(w);
	}
	/* Now move the spaces around randomly */
	if (w->triangles.size > 1) {
		int         big = w->triangles.sizeSize + NRAND(2);
		int         lastDirection = -1;
		int         randomDirection;

		cb.reason = TRIANGLES_RESET;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);

		if (w->triangles.size == 2)
			big *= big;

#ifdef DEBUG
		big = 3;
#endif

		if (big > 1000)
			big = 1000;
		while (big--) {
			randomDirection = NRAND(COORD);

#ifdef DEBUG
			sleep(1);
#endif

			if ((randomDirection + COORD / 2) % COORD != lastDirection) {
				if (MoveTrianglesDir(w, randomDirection, INSTANT))
					lastDirection = randomDirection;
				else
					big++;
			}
		}
		FlushMoves(w);
		cb.reason = TRIANGLES_RANDOMIZE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	if (CheckSolved(w)) {
		cb.reason = TRIANGLES_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
SolveTiles(TrianglesWidget w)
{
	if (CheckSolved(w) || w->triangles.currentPosition >= 0)
		return;
	{
		trianglesCallbackStruct cb;

		cb.reason = TRIANGLES_SOLVE_MESSAGE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
IncrementTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	cb.reason = TRIANGLES_INC;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
DecrementTiles(TrianglesWidget w)
{
	trianglesCallbackStruct cb;

	if (w->triangles.size <= MINTRIANGLES)
		return;
	cb.reason = TRIANGLES_DEC;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
MoveTiles(TrianglesWidget w, int from, int orient, int fast)
{
#ifdef ANIMATE
	if (fast != INSTANT && w->triangles.moveDelay > 0) {
		w->triangles.currentPosition = from;
		w->triangles.currentRow[ROW] = Row(from);
		w->triangles.currentRow[TRBL] =
			TrBl(from, w->triangles.currentRow[ROW]);
		w->triangles.currentRow[TLBR] =
			TlBr(from, w->triangles.currentRow[ROW]);
		w->triangles.currentPositionOrient = ToOrient(w->triangles.currentRow[ROW],
			w->triangles.currentRow[TRBL],
			w->triangles.currentRow[TLBR]);

		if (w->triangles.currentPosition < 
			w->triangles.spacePosition[w->triangles.currentPositionOrient]) {
			AnimateSlide(w, DOWN, fast, False);
		} else if (w->triangles.currentPosition > 
			w->triangles.spacePosition[w->triangles.currentPositionOrient]) {
			AnimateSlide(w, UP, fast, False);
		} else
			return;
		w->triangles.currentPosition = -1;
	} else
#endif
	{
		int         tempTile;

		tempTile = w->triangles.tileOfPosition[from];
		w->triangles.tileOfPosition[from] =
			w->triangles.tileOfPosition[w->triangles.spacePosition[orient]];
		w->triangles.tileOfPosition[w->triangles.spacePosition[orient]] =
			tempTile;
		DrawTile(w, w->triangles.spacePosition[orient], orient, False, False, FALSE, 0, 0);
		w->triangles.spacePosition[orient] = from;
		w->triangles.spaceRow[orient][ROW] = Row(from);
		w->triangles.spaceRow[orient][TRBL] =
			TrBl(from, w->triangles.spaceRow[orient][ROW]);
		w->triangles.spaceRow[orient][TLBR] =
			TlBr(from, w->triangles.spaceRow[orient][ROW]);
		DrawTile(w, w->triangles.spacePosition[orient], orient, True, True, FALSE, 0, 0);
	}
}

static int
ExchangeTiles(TrianglesWidget w, int pos1, int pos2)
{
	int         tempTile;

	if (w->triangles.tileOfPosition[pos1] <= 0)
		return FALSE;
	else if (w->triangles.tileOfPosition[pos2] <= 0)
		return FALSE;
	tempTile = w->triangles.tileOfPosition[pos1];
	w->triangles.tileOfPosition[pos1] = w->triangles.tileOfPosition[pos2];
	w->triangles.tileOfPosition[pos2] = tempTile;
	return TRUE;
}

static int
TileNFromSpace(TrianglesWidget w, int n, int rowType, int orient, int direction)
{
	int pos = w->triangles.spacePosition[orient];

	if (rowType == ROW) /* This one is easy */
		return (pos + ((direction == UP) ? -n : n));
	else {
		int row = Row(pos);
		int trbl = TrBl(pos, row);
		int tlbr = TlBr(pos, row);
		int offset1, offset2; 

		if (direction == UP) {
			offset1 = -(n / 2 + (n & 1));
			offset2 = -(n / 2); 
		} else {
			offset1 = n / 2; 
			offset2 = n / 2 + (n & 1);
		}
		if (rowType == TRBL)
			return ((orient == DOWN) ?
			  ToPosition(row + offset1, trbl, tlbr + offset2) :
			  ToPosition(row + offset2, trbl, tlbr + offset1));
		else if (rowType == TLBR)
			return ((orient == DOWN) ?
			  ToPosition(row + offset1, trbl + offset2, tlbr) :
			  ToPosition(row + offset2, trbl + offset1, tlbr));
		else	/* rowType == ROW */
			return (pos + ((direction == UP) ? -n : n));
	}
}

static void
DrawFrame(TrianglesWidget w, GC gc)
{
	int         sumX, sumY, sumX2, offsetX, offsetY;
	XPoint      tempList[5];

	offsetX = w->triangles.puzzleOffset.x;
	offsetY = w->triangles.puzzleOffset.y;
	sumX = w->triangles.size * w->triangles.offset.x +
		w->triangles.delta.x + 1;
	sumY = w->triangles.size * w->triangles.offset.y +
		w->triangles.delta.y + 1;
	sumX2 = sumX / 2 + offsetX;
	sumX += offsetX;
	sumY += offsetY;
	if (offsetY > 0)
		XFillRectangle(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
			0, 0, w->core.width, offsetY);
	if (sumY < w->core.height)
		XFillRectangle(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
			0, sumY + 1, w->core.width, w->core.height - sumY);
	if (offsetX > 0)
		XFillRectangle(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
			0, 0, offsetX, w->core.height);
	if (sumX - 2 < w->core.width)
		XFillRectangle(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
			sumX - 2, 0, w->core.width - sumX + 2, w->core.height);
	tempList[0].x = offsetX;
	tempList[0].y = offsetY;
	tempList[1].x = sumX2 - 1;
	tempList[1].y = offsetY;
	tempList[2].x = offsetX;
	tempList[2].y = sumY;
	XFillPolygon(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
		tempList, 3, Convex, CoordModeOrigin);
	tempList[0].x = sumX - 2;
	tempList[0].y = offsetY;
	tempList[1].x = sumX - 2;
	tempList[1].y = sumY;
	tempList[2].x = sumX2 - 2;
	tempList[2].y = offsetY;
	XFillPolygon(XtDisplay(w), XtWindow(w), w->triangles.borderGC,
		tempList, 3, Convex, CoordModeOrigin);
	tempList[0].x = sumX2 - 2;
	tempList[0].y = offsetY;
	tempList[1].x = sumX - 2;
	tempList[1].y = sumY;
	tempList[2].x = offsetX;
	tempList[2].y = sumY;
	tempList[3].x = sumX2 - 1;
	tempList[3].y = offsetY;
	tempList[4].x = sumX2 - 2;
	tempList[4].y = offsetY;
	XDrawLines(XtDisplay(w), XtWindow(w), gc,
		tempList, 5, CoordModeOrigin);
}

void
DrawAllTiles(TrianglesWidget w)
{
	int         k, side = UP, fin = 1, step = 1;

	for (k = 0; k < w->triangles.sizeSize; k++) {
		DrawTile(w, k, side, (w->triangles.tileOfPosition[k] <= 0), (w->triangles.tileOfPosition[k] <= 0), FALSE, 0, 0);
		if (fin == k + 1) {
			side = UP;
			step += 2;
			fin += step;
		} else
			side = !side;
	}
}

static void
drawShadow(TrianglesWidget w, GC gc, int startX, int startY,
	XPoint * list, Boolean orient)
{
	XPoint tempList[3];
	int i;

	for (i = 0; i < 3; i++) {
		tempList[i].x = ((i == 0) ? startX : tempList[i - 1].x) + list[i].x;
		tempList[i].y = ((i == 0) ? startY : tempList[i - 1].y) + list[i].y;
	}
	if (orient) {
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			tempList[2].x, tempList[2].y,
			tempList[0].x, tempList[0].y);
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			tempList[2].x + 1, tempList[2].y,
			tempList[0].x, tempList[0].y + 1);
	} else {
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			tempList[0].x, tempList[0].y,
			tempList[1].x, tempList[1].y);
		XDrawLine(XtDisplay(w), XtWindow(w), gc,
			tempList[1].x, tempList[1].y,
			tempList[2].x, tempList[2].y);
	}
}

static void
fill3DTriangle(TrianglesWidget w, GC gc, GC darkerGC, GC brighterGC,
	XPoint * list, Boolean raised, Boolean orient)
{
	GC currentGC = (raised) ? gc : darkerGC;
	XPoint tempList[3];
	int i;

	for (i = 0; i < 3; i++) {
		tempList[i].x = ((i == 0) ? 0 : tempList[i - 1].x) + list[i].x;
		tempList[i].y = ((i == 0) ? 0 : tempList[i - 1].y) + list[i].y;
	}

	XFillPolygon(XtDisplay(w), XtWindow(w), currentGC, list, 3,
		Convex, CoordModePrevious);
	currentGC = (raised ^ orient) ? ((raised) ? brighterGC : gc) :
		darkerGC;
	XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
		tempList[0].x, tempList[0].y, tempList[1].x, tempList[1].y);
	XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
		tempList[1].x, tempList[1].y, tempList[2].x, tempList[2].y);
	if (orient) {
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[0].x, tempList[0].y + 1,
			tempList[1].x - 1, tempList[1].y);
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[1].x - 1, tempList[1].y - 1,
			tempList[2].x + 1, tempList[2].y - 1);
	} else {
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[0].x, tempList[0].y - 1,
			tempList[1].x + 1, tempList[1].y);
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[1].x + 1, tempList[1].y + 1,
			tempList[2].x - 1, tempList[2].y + 1);
	}
	currentGC = (raised ^ orient) ? darkerGC :
		((raised) ? brighterGC : gc);
	XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
		tempList[2].x, tempList[2].y, tempList[0].x, tempList[0].y);
	if (orient) {
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[2].x + 1, tempList[2].y,
			tempList[0].x, tempList[0].y + 1);
	} else {
		XDrawLine(XtDisplay(w), XtWindow(w), currentGC,
			tempList[2].x - 1, tempList[2].y,
			tempList[0].x, tempList[0].y - 1);
	}
}

static void
DrawTile(TrianglesWidget w, int pos, int orient, Boolean blank, Boolean erase, int pressedOffset, int offsetX, int offsetY)
{
	int         dx, dy;
	GC          tileGC, borderGC;

	if (erase) {
		tileGC = w->triangles.inverseGC;
		borderGC = w->triangles.inverseGC;
	} else {
		tileGC = w->triangles.tileGC;
		borderGC = w->triangles.borderGC;
	}
	Cartesian(w, pos, orient, &dx, &dy);
	dx += offsetX + pressedOffset;
	dy += offsetY + pressedOffset;
	triangleList[orient][0].x = dx;
	triangleList[orient][0].y = dy;
	if (blank) {
		XFillPolygon(XtDisplay(w), XtWindow(w), tileGC,
			triangleList[orient], 3,
			Convex, CoordModePrevious);
		XDrawLines(XtDisplay(w), XtWindow(w), borderGC,
			triangleList[orient], 4,
			CoordModePrevious);
	} else {
		if (pressedOffset != 0) {
			drawShadow(w, w->triangles.tileDarkerGC,
				-pressedOffset, -pressedOffset,
				triangleList[orient], orient == UP);
		}
		fill3DTriangle(w, tileGC, w->triangles.tileDarkerGC,
			w->triangles.tileBrighterGC,
			triangleList[orient],
			pressedOffset == 0, orient == UP);
	}
	if (!blank) {
		int         i = 0, digitOffsetX = 0, digitOffsetY = 0;
		int         tile = w->triangles.tileOfPosition[pos];
		char        buf[65];

		(void) int2String(buf, tile, w->triangles.base);
		while (tile >= 1) {
			tile /= w->triangles.base;
			digitOffsetX += w->triangles.digitOffset.x;
			i++;
		}
		digitOffsetY = (orient == UP) ? w->triangles.digitOffset.y +
			w->triangles.tileSize.y / 8 +
			2 * w->triangles.delta.y :
			-w->triangles.digitOffset.y - w->triangles.tileSize.y +
			- w->triangles.tileSize.y / 7 +
			w->triangles.delta.y + 2;
		(void) XDrawString(XtDisplay(w), XtWindow(w), borderGC,
			dx - digitOffsetX,
			dy + w->triangles.tileSize.y / 2 + digitOffsetY,
			buf, i);
	}
}

#ifdef RHOMBUS
static void
DrawRhombus(TrianglesWidget w, int pos, int rowType, int offsetX, int offsetY)
{
	int         dx, dy, k = Row(pos);

	dy = k * w->triangles.offset.y + w->triangles.delta.y + 2 +
		w->triangles.puzzleOffset.y - 2;
	dx = (pos - k * k - k + w->triangles.size) * w->triangles.offset.x / 2 +
		w->triangles.delta.x / 2 + w->triangles.puzzleOffset.x - 1;
	dx += offsetX;
	dy += offsetY;
	rhombusList[rowType][0].x = dx;
	rhombusList[rowType][0].y = dy;
	XFillPolygon(XtDisplay(w), XtWindow(w), w->triangles.inverseGC,
		rhombusList[rowType], 4,
		Convex, CoordModePrevious);
	XDrawLines(XtDisplay(w), XtWindow(w), w->triangles.inverseGC,
		rhombusList[rowType], 5,
		CoordModePrevious);
}
#endif

static void
DiscreteMoves(TrianglesWidget w, int dir, int fast)
{
        trianglesCallbackStruct cb;
	int rowType, orient, next;
	char *buf;

	rowType = MovableTile(w);
	if (rowType < 0) {
		intCat(&buf, "DiscreteMoves: rowType ", rowType);
		XtWarning(buf);
		free(buf);
		XtWarning(buf);
	}
	if (dir == DOWN) {
		orient = (w->triangles.spacePosition[UP] + (rowType == TRBL) <
			w->triangles.spacePosition[DOWN]) ? UP : DOWN;
		next = TileNFromSpace(w, 1, rowType, orient, UP);
		orient = !orient;
		MoveTiles(w, next, orient, fast);
		cb.reason = TRIANGLES_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		switch (rowType) {
			case TLBR:
				PutMove(BR);
				break;
			case TRBL:
				PutMove(BL);
				break;
			case ROW:
				PutMove(RIGHT);
				break;
			default:
				{
					intCat(&buf,
						"DiscreteMoves: rowType ",
						rowType);
					XtWarning(buf);
					free(buf);
					return;
				}
		}
	} else {
		orient = (w->triangles.spacePosition[UP] + 2 * (rowType == TRBL) >
			w->triangles.spacePosition[DOWN]) ? UP : DOWN;
		next = TileNFromSpace(w, 1, rowType, orient, DOWN);
		orient = !orient;
		MoveTiles(w, next, orient, fast);
		cb.reason = TRIANGLES_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		switch (rowType) {
			case TLBR:
				PutMove(TL);
				break;
			case TRBL:
				PutMove(TR);
				break;
			case ROW:
				PutMove(LEFT);
				break;
			default:
				{
					intCat(&buf,
						"DiscreteMoves: rowType ",
						rowType);
					XtWarning(buf);
					free(buf);
					return;
				}
		}
	}
}

#ifdef ANIMATE
static int
CountTiles(TrianglesWidget w, int *dir, int *rowType)
{
	int count = 0;
	char *buf;

	*rowType = MovableTile(w);
	if (*rowType < 0) {
		intCat(&buf, "CountTiles: rowType ", *rowType);
		XtWarning(buf);
		free(buf);
		return 0;
	}
	switch (*rowType) {
		case TLBR:
			count = w->triangles.spaceRow[*dir][TRBL] -
			  w->triangles.currentRow[TRBL] +
			  w->triangles.spaceRow[!*dir][ROW] -
			  w->triangles.currentRow[ROW];
			*dir = (*dir == DOWN) ? BR : TL;
			break;
		case TRBL:
			count = w->triangles.spaceRow[*dir][TLBR] -
			  w->triangles.currentRow[TLBR] +
			  w->triangles.spaceRow[!*dir][ROW] -
			  w->triangles.currentRow[ROW];
			*dir = (*dir == DOWN) ? BL: TR;
			break;
		case ROW:
			count = w->triangles.spaceRow[*dir][TRBL] -
			  w->triangles.currentRow[TRBL] +
			  w->triangles.currentRow[TLBR] - 
			  w->triangles.spaceRow[!*dir][TLBR];
			*dir = (*dir == DOWN) ? RIGHT: LEFT;
			break;
		default:
			{
				intCat(&buf, "CountTiles: rowType ",
					*rowType);
				XtWarning(buf);
				free(buf);
			}
	}
	return count;
}

static void
AnimateSlide(TrianglesWidget w, int dir, int fast, Bool logMoves)
{
        trianglesCallbackStruct cb;
        int aTile, numTiles, direction = dir, rowType;
	int fillBeginPos, fillBeginPos2, fillBeginOrient, fillBeginRow;
#if 0
	int fillBeginI, fillBeginJ;
#endif
	int pos, pos2, posNext, orient, orientNext;

	numTiles = CountTiles(w, &direction, &rowType);
	if (numTiles < 0)
		numTiles = -numTiles;
	if (dir == DOWN)
	  orient = (w->triangles.spacePosition[UP] + (rowType == TRBL) <
	    w->triangles.spacePosition[DOWN]) ? UP : DOWN;
	else
	  orient = (w->triangles.spacePosition[UP] + 2 * (rowType == TRBL) >
	    w->triangles.spacePosition[DOWN]) ? UP : DOWN;
	fillBeginPos = TileNFromSpace(w, numTiles, rowType, orient, !dir);
	fillBeginRow = Row(fillBeginPos);
	fillBeginOrient = ToOrient(fillBeginRow,
		TrBl(fillBeginPos, fillBeginRow),
		TlBr(fillBeginPos, fillBeginRow));
	fillBeginPos2 = w->triangles.spacePosition[!fillBeginOrient];
#if 0
	fillBeginJ = (fillBeginPos == UP) ? 0 : w->triangles.tileSize.y;
	fillBeginJ += fillBeginRow * w->triangles.offset.y +
		w->triangles.delta.y + 2 +
		w->triangles.puzzleOffset.y; 
	fillBeginI = (fillBeginPos - fillBeginRow * fillBeginRow -
		fillBeginRow + w->triangles.size) * w->triangles.offset.x / 2 +
		w->triangles.delta.x / 2 + w->triangles.puzzleOffset.x;
#endif
#ifdef RHOMBUS
	{
		int inc = 0;
        	int gapI = 0, moveI = 0;
        	int gapJ = 0, moveJ = 0;

		gapI = w->triangles.tileSize.x * fast / w->triangles.numSlices;
		moveI = w->triangles.tileSize.x + w->triangles.delta.x;
		gapJ = 0;
		moveJ = w->triangles.tileSize.y + w->triangles.delta.y;
		if (gapI == 0)
		  gapI++;	
		for (inc = 0; inc < moveI + gapI; inc += gapI) {
		  if (inc > moveI)
		    inc = moveI;
		  for (aTile = 0; aTile < numTiles; aTile++) {
		    posNext = TileNFromSpace(w, aTile + 1, rowType, orient, !dir);
		    orientNext = (aTile & 1) ? orient : !orient;
		    /* Calculate deltas */
		    DrawTile(w, posNext, orientNext, False, False, FALSE, 1, 1);
		    /* Erase old slivers */
		  }
		  XFlush(XtDisplay(w));
		  Sleep(w->triangles.moveDelay / fast);
	}
#else
	{
		int pixNo = 0;
		int pixX = 0;
		int pixY = 0;
		int fromdx = 0;
		int fromdy = 0;
		int todx = 0;
		int tody = 0;
		int dx, dy;
		int firstRow;
		int *rowPos;
		int i, k;
		Boolean slideDone = False;
		int firstOrient;

		if (!(rowPos = (int *)
		      malloc(sizeof (int) * (numTiles + 2))))
		            XtError("Not enough memory, exiting.");
		rowPos[2] = TileNFromSpace(w, 1, rowType, orient, !dir);
		firstRow = Row(rowPos[2]);
		firstOrient = ToOrient(firstRow,
			TrBl(rowPos[2], firstRow),
			TlBr(rowPos[2], firstRow));
		rowPos[0] = w->triangles.spacePosition[firstOrient];
		rowPos[1] = w->triangles.spacePosition[!firstOrient];
		for (i = 1; i < numTiles; i++) {
			rowPos[i + 2] = TileNFromSpace(w,
				i + 1, rowType, orient, !dir);
		}
		Cartesian(w, rowPos[2], firstOrient, &fromdx, &fromdy);
		Cartesian(w, rowPos[0], firstOrient, &todx, &tody);
		dx = todx - fromdx;
		dy = tody - fromdy;
		i = 0;
		while (!slideDone) {
			for (k = 2; k < numTiles + 2; k++)
				DrawTile(w, rowPos[k],
					(firstOrient + k) & 1,
					True, True, FALSE, pixX, pixY);
			slideDone = bresenhamLine(w, 0, 0, dx, dy,
				&pixX, &pixY, &pixNo);
			for (k = 2; k < numTiles + 2; k++)
				DrawTile(w, rowPos[k],
					(firstOrient + k) & 1,
					False, False, FALSE, pixX, pixY);
			XFlush(XtDisplay(w));
			if (i % 8 == 0)
				Sleep(w->triangles.moveDelay / fast);
			i++;
		}
	}
#endif
	for (aTile = 0; aTile < numTiles; aTile++) {
	  if (logMoves) {
	    cb.reason = TRIANGLES_MOVED;
	    XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	    PutMove(direction);
	  }
	  posNext = TileNFromSpace(w, aTile + 1, rowType, orient, !dir);
	  orientNext = (aTile & 1) ? orient : !orient;
	  if (orientNext == !fillBeginOrient)
	    fillBeginPos2 = posNext;
	}
	pos2 = w->triangles.spacePosition[orient];
	pos = w->triangles.spacePosition[!orient];
	for (aTile = 0; aTile < numTiles; aTile++) {
	  posNext = TileNFromSpace(w, aTile + 1, rowType, orient, !dir);
          w->triangles.tileOfPosition[pos] =
            w->triangles.tileOfPosition[posNext];
          pos = pos2;
          pos2 = posNext;
	}
        w->triangles.tileOfPosition[fillBeginPos] =
	  w->triangles.spacePosition[fillBeginOrient];
        w->triangles.tileOfPosition[fillBeginPos2] =
	  w->triangles.spacePosition[!fillBeginOrient];
	w->triangles.spacePosition[fillBeginOrient] = fillBeginPos;
	w->triangles.spacePosition[!fillBeginOrient] = fillBeginPos2;
	if (fillBeginOrient == UP) {
	  w->triangles.tileOfPosition[fillBeginPos] = 0;
	  w->triangles.tileOfPosition[fillBeginPos2] = -1;
	} else {
	  w->triangles.tileOfPosition[fillBeginPos] = -1;
	  w->triangles.tileOfPosition[fillBeginPos2] = 0;
	}
	w->triangles.spaceRow[fillBeginOrient][ROW] = Row(fillBeginPos);
	w->triangles.spaceRow[fillBeginOrient][TRBL] =
	  TrBl(fillBeginPos, w->triangles.spaceRow[fillBeginOrient][ROW]);
	w->triangles.spaceRow[fillBeginOrient][TLBR] =
	  TlBr(fillBeginPos, w->triangles.spaceRow[fillBeginOrient][ROW]);
	w->triangles.spaceRow[!fillBeginOrient][ROW] = Row(fillBeginPos2);
	w->triangles.spaceRow[!fillBeginOrient][TRBL] =
	  TrBl(fillBeginPos2, w->triangles.spaceRow[!fillBeginOrient][ROW]);
	w->triangles.spaceRow[!fillBeginOrient][TLBR] =
	  TlBr(fillBeginPos2, w->triangles.spaceRow[!fillBeginOrient][ROW]);
#ifdef RHOMBUS
	DrawRhombus(w, w->triangles.spacePosition[UP], RhombusType(w), 1, 1);
#endif
}
#endif

static Boolean
bresenhamLine(TrianglesWidget w, int x0, int y0, int x1, int y1, 
	int * pixX, int * pixY, int * pixNo)
{
	int pix = 0;
	int ex = x1 - x0;
	int ey = y1 - y0;
	int dx, dy, error;

	(*pixNo)++;
	if (ex > 0) {
		dx = 1;
	} else if (ex < 0) {
		dx = -1;
		ex = -ex;
	} else {
		dx = 0;
	}
	if (ey > 0) {
		dy = 1;
	} else if (ey < 0) {
		dy = -1;
		ey = -ey;
	} else {
		dy = 0;
	}
	*pixX = x0;
	*pixY = y0;
	if (ex > ey) {
		error = 2 * ey - ex;
		while (pix != *pixNo) {
			if (error >= 0) {
				error -= 2 * ex;
				*pixY += dy;
			}
			error += 2 * ey;
			*pixX += dx;
			pix++;
			if (*pixX == x1)
				return True;
		}
		return False;
	} else {
		error = 2 * ex -ey;
		while (pix != *pixNo) {
			if (error >= 0) {
				error -= 2 * ey;
				*pixX += dx;
			}
			error += 2 * ex;
			*pixY += dy;
			pix++;
			if (*pixY == y1)
				return True;
		}
		return False;
	}
}

static void
Cartesian(TrianglesWidget w, int pos, int orient, int * x, int * y)
{
	int k = Row(pos);

	*y = (orient == UP) ? 0 : w->triangles.tileSize.y;
	*y += k * w->triangles.offset.y + w->triangles.delta.y + 2 +
		w->triangles.puzzleOffset.y;
	*x = (pos - k * k - k + w->triangles.size) * w->triangles.offset.x / 2 +
		w->triangles.delta.x / 2 + w->triangles.puzzleOffset.x;
}

int
Row(int pos)
{
	return Sqrt(pos);
}

/* Passing row so there is no sqrt calculation again */
int
TrBl(int pos, int posRow)
{
	return ((pos - posRow * posRow) / 2);
}

int
TlBr(int pos, int posRow)
{
	return ((posRow * posRow + 2 * posRow - pos) / 2);
}

static int
ToOrient(int row, int trbl, int tlbr)
{
	return (trbl + tlbr == row);
}

static int
ToPosition(int row, int trbl, int tlbr)
{
	return (row * row + row + trbl - tlbr);
}

/* This is fast for small i, a -1 is returned for negative i. */
static int
Sqrt(int i)
{
	int         j = 0;

	while (j * j <= i)
		j++;
	return (j - 1);
}

static int
int2String(char *buf, int number, int base)
{
	int         digit, mult = base, last, position;

	if (number < 0) {
		char *buff;

		intCat(&buff, "int2String: 0 > ", number);
		XtWarning(buff);
		free(buff);
		return 0;
	}
	last = 1;
	while (number >= mult) {
		last++;
		mult *= base;
	}
	for (position = 0; position < last; position++) {
		mult /= base;
		digit = number / mult;
		number -= digit * mult;
		buf[position] = digit + '0';
		if (buf[position] > '9') {	/* ASCII */
			buf[position] += ('A' - '9' - 1);
		} else if (buf[position] < '0') {	/* EBCDIC */
			buf[position] += ('A' - '9' - 1);
			if (buf[position] > 'I')
				buf[position] += ('J' - 'I' - 1);
			if (buf[position] > 'R')
				buf[position] += ('S' - 'R' - 1);
		}
	}
	buf[last] = '\0';
	return last;
}

Boolean
CheckSolved(TrianglesWidget w)
{
	int         i;

	for (i = 1; i < w->triangles.sizeSize - 1; i++)
		if (w->triangles.tileOfPosition[i - 1] != i)
			return FALSE;
	return TRUE;
}
